"""REST API Module with automatic OpenAPI generation."""
import asyncio
import contextlib
import os
from functools import wraps
from inspect import signature
from typing import Any, Awaitable, Callable, TypeVar, cast

from aiohttp.typedefs import Handler
from aiohttp.web import Application, HTTPFound, Request, Response, StreamResponse
from aiohttp.web_middlewares import _Middleware as Middleware  # type: ignore
from aiohttp.web_middlewares import middleware
from aiohttp.web_ws import WebSocketResponse
from aiohttp_sse import EventSourceResponse, sse_response  # type: ignore
from pydantic import BaseModel  # pylint: disable=no-name-in-module
from typing_extensions import ParamSpec

from .docs import extract, html_string, load, transform
from .helpers import do_response
from .json import jsonable_encoder
from .utils import setup_logging

P = ParamSpec("P")
T = TypeVar("T")
S = TypeVar("S", bound=StreamResponse)


class APIServer(Application):
    """Cloudantic APIServer"""

    def __init__(
        self,
        *args: Any,
        title: str = "Cloudantic",
        servers: list[str] | None = None,
        description: str = "Cloudantic API",
        version: str = "0.0.1",
        openapi_url: str = "/openapi.json",
        **kwargs: Any,
    ):
        super().__init__(*args, logger=setup_logging(self.__class__.__name__), **kwargs)
        schemas = BaseModel.__subclasses__()
        self.openapi: dict[str, Any] = {
            "openapi": "3.0.0",
            "info": {"title": title, "version": version},
            "paths": {},
            "tags": [],
            "components": {
                "schemas": {schema.__name__: schema.schema() for schema in schemas}
            },
            "description": description,
        }
        if servers:
            self.openapi["servers"] = servers
        self._route_open_api_params: dict[str, Any] = {}
        self.openapi_url = openapi_url

        @self.get("/openapi.json")
        async def _():
            response = jsonable_encoder(self.openapi)
            return response

        @self.get("/docs")
        async def _():
            return Response(
                text=html_string(self.openapi_url), content_type="text/html"
            )

    def document(
        self, path: str, method: str
    ) -> Callable[[Callable[P, Awaitable[S]]], Callable[P, Awaitable[S]]]:
        """

        SWAGGER DOCUMENTATION

        """

        def decorator(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
            sig = signature(func)
            params = sig.parameters
            open_api_params = extract(params.copy(), path)
            self._route_open_api_params[(path, method)] = open_api_params  # type: ignore
            transform(self.openapi, path, method, func, open_api_params)

            async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                request: Request = cast(Request, args[0])
                args_ = args[1:]
                args_to_apply = await load(request, params.copy())
                definitive_args = {}
                for name, param in params.items():
                    if name in args_to_apply:
                        definitive_args[name] = args_to_apply[name]
                    elif param.default is not param.empty:
                        definitive_args[name] = param.default
                    else:
                        raise ValueError(
                            f"Missing parameter {name} for {func.__name__}"
                        )
                if asyncio.iscoroutinefunction(func):
                    response = await func(*args_, **kwargs, **definitive_args)  # type: ignore
                else:
                    response = func(*args_, **kwargs, **definitive_args)  # type: ignore
                return cast(T, do_response(response))

            wrapper._handler = func  # type: ignore
            return wrapper

        return decorator

    def get(  # type: ignore
        self, path: str, **kwargs: Any
    ) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
        """GET decorator"""

        def decorator(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
            self.router.add_get(path, self.document(path, "GET")(func), **kwargs)  # type: ignore
            return func

        return decorator

    def post(
        self, path: str, **kwargs: Any
    ) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
        """POST decorator"""

        def decorator(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
            self.router.add_post(path, self.document(path, "POST")(func), **kwargs)  # type: ignore
            return func

        return decorator

    def put(
        self, path: str, **kwargs: Any
    ) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
        """PUT decorator"""

        def decorator(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
            self.router.add_put(path, self.document(path, "PUT")(func), **kwargs)  # type: ignore
            return func

        return decorator

    def delete(
        self, path: str, **kwargs: Any
    ) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
        """DELETE decorator"""

        def decorator(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
            self.router.add_delete(path, self.document(path, "DELETE")(func), **kwargs)  # type: ignore
            return func

        return decorator

    def patch(
        self, path: str, **kwargs: Any
    ) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
        """PATCH decorator"""

        def decorator(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
            self.router.add_patch(path, self.document(path, "PATCH")(func), **kwargs)  # type: ignore
            return func

        return decorator

    def head(
        self, path: str, **kwargs: Any
    ) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
        """HEAD decorator"""

        def decorator(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
            self.router.add_head(path, self.document(path, "HEAD")(func), **kwargs)  # type: ignore
            return func

        return decorator

    def options(self, path: str, **kwargs: Any):
        """OPTIONS decorator"""

        def decorator(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
            self.router.add_options(
                path, self.document(path, "OPTIONS")(func), **kwargs  # type: ignore
            )  # type: ignore
            return func

        return decorator

    def on_event(
        self, event: str
    ) -> Callable[
        [Callable[[Application], Awaitable[None]]],
        Callable[[Application], Awaitable[None]],
    ]:
        """On event handler"""

        def decorator(
            func: Callable[[Application], Awaitable[None]]
        ) -> Callable[[Application], Awaitable[None]]:
            if event not in ("startup", "shutdown"):
                raise ValueError("Event must be startup or shutdown")
            elif event == "startup":
                self.on_startup.append(func)
            else:
                self.on_shutdown.append(func)
            return func

        return decorator

    def sse(
        self, path: str
    ) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]],]:
        """Server Sent Events decorator"""

        def decorator(func: Callable[P, Awaitable[S]]) -> Callable[P, Awaitable[S]]:
            @wraps(func)
            async def wrapper(*args: P.args, **kwargs: P.kwargs) -> S:
                request: Request = cast(Request, args[0])
                async with sse_response(request) as resp:  # type: ignore
                    args_to_apply = await load(
                        request, signature(func).parameters.copy()
                    )
                    definitive_args = {}
                    for name, param in signature(func).parameters.items():
                        if param.annotation == EventSourceResponse:
                            definitive_args[name] = resp  # type: ignore
                        elif name in args_to_apply:
                            definitive_args[name] = args_to_apply[name]
                            args_to_apply.pop(name)
                        elif param.default is not param.empty:
                            definitive_args[name] = param.default
                        else:
                            raise ValueError(
                                f"Missing parameter {name} for {func.__name__}"
                            )
                    await func(**definitive_args)  # type: ignore
                    return resp  # type: ignore

            self.router.add_get(path, wrapper)
            return wrapper  # type: ignore

        return decorator  # type: ignore

    def websocket(
        self, path: str
    ) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
        """Websocket decorator"""

        def decorator(func: Callable[P, Awaitable[S]]) -> Callable[P, Awaitable[S]]:
            @wraps(func)
            async def wrapper(request: Request):
                args_to_apply = await load(request, signature(func).parameters.copy())
                ws = WebSocketResponse()
                await ws.prepare(request)
                definitive_args = {}
                for name, param in signature(func).parameters.items():
                    if param.annotation == WebSocketResponse:
                        definitive_args[name] = ws
                    elif name in args_to_apply:
                        definitive_args[name] = args_to_apply[name]
                        args_to_apply.pop(name)
                    elif param.default is not param.empty:
                        definitive_args[name] = param.default
                    else:
                        raise ValueError(
                            f"Missing parameter {name} for {func.__name__}"
                        )
                await func(**definitive_args)  # type: ignore
                return ws

            self.router.add_get(path, wrapper)
            return wrapper  # type: ignore

        return decorator  # type: ignore

    def static(self, prefix: str, path: str) -> "APIServer":
        """Static folder creation and serving"""
        with contextlib.suppress(OSError):
            os.makedirs(path)
        self.router.add_static(prefix, path)

        return self

    def middleware(self, func: Middleware) -> Middleware:
        @wraps(func)
        @middleware
        async def wrapper(request: Request, handler: Handler) -> Response:
            response = await func(request, handler)
            if isinstance(response, Response):
                return response
            return do_response(response)

        self.middlewares.append(wrapper)
        return wrapper



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Literal, TypeAlias, TypeVar, Union

from pydantic import BaseModel  # pylint: disable=no-name-in-module
from typing_extensions import ParamSpec

ImageModel: TypeAlias = Literal["dall-e"]
CompletionModel: TypeAlias = Literal["gpt-3.5-turbo-instruct", "davinci-002"]
ChatModel: TypeAlias = Literal["gpt-4", "gpt-3.5-turbo", "gpt-3.5-turbo-16k"]
EmbeddingModel: TypeAlias = Literal["text-embedding-ada-002"]
AudioModel: TypeAlias = Literal["whisper-1"]
Model: TypeAlias = Union[
    ChatModel, EmbeddingModel, AudioModel, CompletionModel, ImageModel
]
Role: TypeAlias = Literal["user", "system", "assistant", "function"]
Size: TypeAlias = Literal["256x256", "512x512", "1024x1024"]
ImageFormat: TypeAlias = Literal["url", "base64"]
AudioFormat: TypeAlias = Literal["mp3", "mp4", "mpeg", "mpga", "m4a", "wav", "webm"]
Vector: TypeAlias = List[float]
MetaDataValue: TypeAlias = Union[str, int, float, bool, List[str]]
MetaData: TypeAlias = Dict[str, MetaDataValue]

M = TypeVar("M", bound=Model)
P = ParamSpec("P")


class OpenAIResource(ABC, BaseModel):
    model: M  # type: ignore

    @abstractmethod
    async def run(self, *args: Any, **kwargs: Any) -> Any:
        ...



from .functions import *
from .openai import *
from .schema import *
from .server import *
from .service import *
from .typedef import *
from .utils import *


class Agent(BaseModel):
    """
    First class interface for interacting with openai large scale language models.
    """

    namespace: str = Field(...)
    chat_completion: ChatGPT = Field(default_factory=ChatGPT)
    completion: DaVinci = Field(default_factory=DaVinci)
    embeddings: AdaEmbeddings = Field(default_factory=AdaEmbeddings)
    audio: Whisper = Field(default_factory=Whisper)
    image: DallE = Field(default_factory=DallE)
    vector: VectorClient = Field(default_factory=VectorClient)

    async def run(self, text: str) -> AsyncGenerator[str, None]:
        """
        Streams the response from chatcompletion using retrieval augmented generation technique
        """
        try:
            string = ""
            knn = await self.vector.search(text, self.namespace)
            context = "Suggestions from the knowledge base:\n\n" + "\n".join(knn)
            async for message in self.chat_completion.stream(text, context):
                string += message
                yield message
            response_vector = (await self.embeddings.run([string]))[0]
            await self.vector.run(
                text=string,
                namespace=self.namespace,
                vector=response_vector,
                action="upsert",
            )
        except Exception as e:
            yield e.__class__.__name__ + ": " + str(e)

    async def __call__(self, text: str):
        """
        Function calling orchestration
        """
        return await function_call(text)



"""
Logging and error handling utilities for the Swagchain API.
"""
import asyncio
import functools
import logging
from time import perf_counter
from typing import Awaitable, Callable, Generator, Sequence, TypeVar, Union, cast

from aiohttp.web_exceptions import HTTPException
from rich.console import Console
from rich.logging import RichHandler
from rich.pretty import install
from rich.traceback import install as ins
from tenacity import retry as retry_
from tenacity import retry_if_exception_type, stop_after_attempt, wait_exponential
from typing_extensions import ParamSpec

EXCEPTIONS = (
    asyncio.TimeoutError,
    ConnectionError,
    ConnectionRefusedError,
    ConnectionResetError,
    TimeoutError,
    UnicodeDecodeError,
    UnicodeEncodeError,
    UnicodeError,
    TypeError,
    ValueError,
    ZeroDivisionError,
    IndexError,
    AttributeError,
    ImportError,
    ModuleNotFoundError,
    NotImplementedError,
    RecursionError,
    OverflowError,
    KeyError,
    Exception,
    HTTPException,
)

T = TypeVar("T")
P = ParamSpec("P")


def setup_logging(name: str) -> logging.Logger:
    """
    Set's up logging using the Rich library for pretty and informative terminal logs.

    Arguments:
    name -- Name for the logger instance. It's best practice to use the name of the module where logger is defined.
    """
    install()
    ins()
    console = Console(record=True, force_terminal=True)
    console_handler = RichHandler(
        console=console,
        show_time=True,
        show_path=True,
        markup=True,
        rich_tracebacks=True,
        tracebacks_show_locals=True,
        tracebacks_extra_lines=2,
        tracebacks_theme="monokai",
        show_level=False,
    )
    console_handler.setFormatter(logging.Formatter("%(message)s"))
    console_handler.setLevel(logging.INFO)
    logging.basicConfig(level=logging.INFO, handlers=[console_handler])
    logger_ = logging.getLogger(name)
    logger_.setLevel(logging.INFO)
    return logger_


logger = setup_logging(__name__)


def process_time(
    func: Callable[P, Union[Awaitable[T], T]]
) -> Callable[P, Awaitable[T]]:
    """
    A decorator to measure the execution time of a coroutine.

    Arguments:
    func -- The coroutine whose execution time is to be measured.
    """

    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        """
        Wrapper function to time the function call.
        """
        start = perf_counter()
        if asyncio.iscoroutinefunction(func):
            result = await func(*args, **kwargs)
        else:
            result = func(*args, **kwargs)
        end = perf_counter()
        logger.info(
            "Time taken to execute %s: %s seconds", wrapper.__name__, end - start
        )
        return result  # type: ignore

    return wrapper


def handle_errors(
    func: Callable[P, Union[Awaitable[T], T]]
) -> Callable[P, Awaitable[T]]:
    """
    A decorator to handle errors in a coroutine.

    Arguments:
    func -- The coroutine whose errors are to be handled.
    """

    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        """
        Wrapper function to handle errors in the function call.
        """
        try:
            if asyncio.iscoroutinefunction(func):
                response = await func(*args, **kwargs)
                logger.info(response)
                return response  # type: ignore
            response = func(*args, **kwargs)
            logger.info(response)
            return response  # type: ignore
        except EXCEPTIONS as exc:
            detail = getattr(exc, "detail", str(exc))
            if detail:
                detail = str(detail).strip()
                logger.error(detail)
            else:
                detail = exc.__class__.__name__  # type: ignore
                logger.error(detail)
            logger.error(str(exc))
            raise HTTPException(reason=detail) from exc

    return wrapper


def chunker(seq: Sequence[T], size: int) -> Generator[Sequence[T], None, None]:
    """
    A generator function that chunks a sequence into smaller sequences of the given size.

    Arguments:
    seq -- The sequence to be chunked.
    size -- The size of the chunks.
    """
    return (seq[pos : pos + size] for pos in range(0, len(seq), size))


def gen_emptystr() -> str:
    """
    A generator function that returns an empty string.
    """
    return cast(str, None)


def handle(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
    """
    A decorator to apply all decorators to a coroutine.

    Arguments:

    func -- The coroutine to decorate.
    """
    return functools.reduce(
        lambda f, g: g(f),  # type: ignore
        [retry(), handle_errors, process_time],
        func,
    )


def async_io(func: Callable[P, T]) -> Callable[P, Awaitable[T]]:
    """
    Decorator to convert an IO bound function to a coroutine by running it in a thread pool.
    """

    @handle
    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        return await asyncio.to_thread(func, *args, **kwargs)

    return wrapper


def async_cpu(func: Callable[P, T]) -> Callable[P, Awaitable[T]]:
    """
    Decorator to convert a CPU bound function to a coroutine by running it in a process pool.
    """

    @handle
    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        return await asyncio.get_running_loop().run_in_executor(
            None, func, *args, **kwargs
        )

    return wrapper


def retry(
    retries: int = 10, wait: int = 1, max_wait: int = 60
) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
    """Wrap an async function with exponential backoff."""

    def decorator(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
        @functools.wraps(func)
        @retry_(
            stop=stop_after_attempt(retries),
            wait=wait_exponential(multiplier=wait, max=max_wait),
            retry=retry_if_exception_type(EXCEPTIONS),
            reraise=True,
        )
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            return await func(*args, **kwargs)  # type: ignore

        return wrapper

    return cast(Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]], decorator)  # type: ignore    # noqa: E501



from abc import ABC
from typing import Any, List, Optional, Type, TypeVar

import openai
from pydantic import BaseModel  # pylint: disable=no-name-in-module
from typing_extensions import ParamSpec

from .utils import handle  # pylint: disable=E0401

T = TypeVar("T")
P = ParamSpec("P")


class FunctionCall(BaseModel):
    """
    Datastructure for rendering function calls
    """

    name: str
    data: Any


class OpenAIFunction(BaseModel, ABC):
    """
    OpenAI Function
    ---------------

    Schema-driven natural language function orchestration engine, powered by OpenAI functions feature and Pydantic.

    It enables the users to easily create and orchestrate functions that can be called from a natural language interface by defining the input schema and implementing the execution logic.

    The schema is defined using Pydantic models, and the execution logic is implemented by overriding the `run` method.

    It's an abstract class, so it can't be instantiated directly, but it can be subclassed to create new functions.

    Example:

    ```python

    import asyncio

    from typing import Any, Optional

    from pydantic import BaseModel

    from liferay_llm import OpenAIFunction

    class AddFunction(OpenAIFunction):
        '''
        Adds two numbers
        '''

        x: int
        y: int

        async def run(self, **kwargs: Any):
            return self.x + self.y

    async def main():
        print(await function_call("add 2 and 3"))

    if __name__ == "__main__":
        asyncio.run(main())

    >>> 5

    ```
    """

    @classmethod
    def __init_subclass__(cls, **kwargs: Any):
        super().__init_subclass__(**kwargs)
        _schema = cls.schema()
        if cls.__doc__ is None:
            raise ValueError(
                f"OpenAIFunction subclass {cls.__name__} must have a docstring"
            )
        cls.openaischema = {
            "name": cls.__name__,
            "description": cls.__doc__,
            "parameters": {
                "type": "object",
                "properties": {
                    k: v for k, v in _schema["properties"].items() if k != "self"
                },
                "required": _schema.get("required", []),
            },
        }

    @handle
    async def __call__(self, **kwargs: Any) -> FunctionCall:
        response = await self.run(**kwargs)
        return FunctionCall(name=self.__class__.__name__, data=response)

    async def run(self, **kwargs: Any) -> Any:
        """Main function to be implemented by subclasses"""
        raise NotImplementedError


@handle
async def parse_openai_function(
    response: dict[str, Any],
    functions: List[Type[OpenAIFunction]] = OpenAIFunction.__subclasses__(),
) -> FunctionCall:
    """Parses the response from the OpenAI API and returns a FunctionCall object, meant to have an uniform interface for rendering the response"""
    choice = response["choices"][0]["message"]
    if "function_call" in choice:
        function_call_ = choice["function_call"]
        name = function_call_["name"]
        arguments = function_call_["arguments"]
        print(name, arguments)
        for i in functions:
            if i.__name__ == name:
                result = await i.parse_raw(arguments)()  # type: ignore
                break
        else:
            raise ValueError(f"Function {name} not found")
        return result
    return FunctionCall(name="chat", data=choice["content"])


@handle
async def function_call(
    text: str,
    model: str = "gpt-3.5-turbo-16k-0613",
    functions: List[Type[OpenAIFunction]] = OpenAIFunction.__subclasses__(),
    **kwargs: Any,
) -> FunctionCall:
    """Given the user input, it will infer the appropiate function to call and it's `json_schema`, then it will call the function and return a FunctionCall object"""
    messages = [
        {"role": "user", "content": text},
        {"role": "system", "content": "You are a function Orchestrator"},
    ]
    response = await openai.ChatCompletion.acreate(  # type: ignore
        model=model,
        messages=messages,
        functions=[func.openaischema for func in functions],
    )
    return await parse_openai_function(response, functions=functions, **kwargs)  # type: ignore


@handle
async def chat_completion(text: str, context: Optional[str] = None) -> str:
    """

    Chat Completion
    ---------------

    Args:

                    text (str): The text to generate from

                    context (Optional[str], optional): The context to generate from. Defaults to None.

    Returns:

                    str: The generated text

    Simple chat completion outside agent context
    """

    if context is not None:
        messages = [
            {"role": "user", "content": text},
            {"role": "system", "content": context},
        ]
    else:
        messages = [{"role": "user", "content": text}]
    response = await openai.ChatCompletion.acreate(  # type: ignore
        model="gpt-3.5-turbo-16k-0613", messages=messages
    )
    return response["choices"][0]["message"]["content"]  # type: ignore


@handle
async def instruction(
    text: str, temperature: float = 0.2, max_tokens: int = 1024
) -> str:
    """

    Instruction Completion
    ---------------

    Args:

                    text (str): The text to generate from

    Returns:

                    str: The generated text

    Simple instruction completion outside agent context
    """
    response = await openai.Completion.acreate(  # type: ignore
        model="gpt-3.5-turbo-instruct",
        prompt=text,
        temperature=temperature,
        max_tokens=max_tokens,
        stream=False,
    )
    return response.choices[0].text  # type: ignore


async def instruction_stream(
    text: str, temperature: float = 0.2, max_tokens: int = 1024
):
    """

    Instruction Completion
    ---------------

    Args:

                    text (str): The text to generate from

    Returns:

                    str: The generated text

    Simple instruction completion outside agent context
    """
    response = await openai.Completion.acreate(  # type: ignore
        model="gpt-3.5-turbo-instruct",
        prompt=text,
        temperature=temperature,
        max_tokens=max_tokens,
        stream=True,
    )
    async for i in response:
        yield i.choices[0].text  # type: ignore



import os
from typing import AsyncGenerator

from pydantic import Field

import openai

from .schema import (
    AudioRequest,
    ChatCompletionRequest,
    ChatCompletionResponse,
    CompletionRequest,
    CompletionResponse,
    Embedding,
    Message,
    Query,
    QueryBuilder,
    QueryRequest,
    QueryResponse,
    UpsertRequest,
    UpsertResponse,
)
from .service import APIClient
from .typedef import (
    AudioFormat,
    AudioModel,
    ChatModel,
    CompletionModel,
    Dict,
    EmbeddingModel,
    ImageFormat,
    ImageModel,
    List,
    Literal,
    OpenAIResource,
    Size,
    Vector,
)
from .utils import handle


class ChatGPT(OpenAIResource):
    """OpenAI Chat Completion API.
    ---
    ChatGPT: A Chat Generative Pretrained Transformer, is a SOAT Large Language model trained on 175 billion parameters with the Reinforcement Learning with Human Feedback technique, a novel invention launched in November's 2022 it's currently enhancing human capabilities around the world in a fast paced way unseen before, you can read more about it at the [OpenAI blog post](https://openai.com/blog/chatgpt).
    """

    model: ChatModel = Field(default="gpt-3.5-turbo-16k")

    @handle
    async def run(self, text: str, context: str):  # type: ignore
        """
        Given an input text and a computed context, decorated with the `handle` utility, this method safely makes calls to the OpenAI Chat Completion API with the appropiate format for an user-llm conversation that has a context window which can store additional contextual information such as memory, similarity retrieval or customized prompts to enforce the behavior of the model for any specific use case. The computation of this context window is up to the user, however sensible defaults such as `retrieval augmented generation` are provided out of the box.

        Args:
            text (str): The user input text.
            context (str): The computed context window.

        Raises:
            NotImplementedError: when an action regarding the retrieval engine which is not implemented is requested.

        Returns:

            ChatCompletionResponse: The response object from the OpenAI Chat Completion API.
        """
        request = ChatCompletionRequest(
            messages=[Message(content=text), Message(content=context, role="system")]
        )
        response = await openai.ChatCompletion.acreate(**request.dict())  # type: ignore
        return ChatCompletionResponse(**response)  # type: ignore

    async def stream(self, text: str, context: str) -> AsyncGenerator[str, None]:
        """
        Similar to the `run` method, but instead of returning a response object, it yields the response chunks of the response message as they are received from the OpenAI Chat Completion API.

        Args:
            text (str): The user input text.
            context (str): The computed context window.

        Raises:
            NotImplementedError: when an action regarding the retrieval engine which is not implemented is requested.

        Yields:
            AsyncGenerator[str, None]: The response chunks of the response message as they are received from the OpenAI Chat Completion API.
        """

        request = ChatCompletionRequest(
            messages=[Message(content=text), Message(content=context, role="system")],
            stream=True,
        )
        response = await openai.ChatCompletion.acreate(**request.dict())  # type: ignore
        async for message in response:  # type: ignore
            data = message.choices[0].delta.get("content", "")  # type: ignore
            yield data  # type: ignore


class DaVinci(OpenAIResource):
    """OpenAI Completion API.
    ---
    Da Vinci is a former version of the GPT-3 model. Da Vinci is a series of text generations models which generate text following the transformer's architecture for a given input with the primary goal of following the input text, it was novel at the time of its release because it was on top of the few text generative models as of today is used for autocompletions, summarizations, translations, etc. and mostly non conversational instruct based tasks, it's drop-in replacement is gpt-3.5-turbo-instruct.
    Read more about it in the [OpenAI blog post](https://openai.com/blog/gpt-3-5-turbo-fine-tuning-and-api-updates).
    """

    model: CompletionModel = Field(default="gpt-3.5-turbo-instruct")

    @handle
    async def run(self, text: str):  # type: ignore
        """
        Given an input text, decorated with the `handle` utility, this method safely makes calls to the OpenAI Completion API with the appropiate format for an user-llm conversation with no context window.
        It is meant for simplier tasks that do not require a context window, such as summarization, translation, autocompletion, etc.

        Args:
            text (str): The user input text.

        Returns:

            CompletionResponse: The response object from the OpenAI Completion API.

        """

        request = CompletionRequest(prompt=text)
        response = await openai.Completion.acreate(**request.dict())  # type: ignore
        return CompletionResponse(**response)  # type: ignore

    async def stream(self, text: str) -> AsyncGenerator[str, None]:
        """

        Similar to the `run` method, but instead of returning a response object, it yields the response chunks of the response message as they are received from the OpenAI Completion API.

        Args:
            text (str): The user input text.

        Yields:
            AsyncGenerator[str, None]: The response chunks of the response message as they are received from the OpenAI Completion API.

        """

        request = CompletionRequest(prompt=text, stream=True)
        response = await openai.Completion.acreate(**request.dict())  # type: ignore
        async for message in response:  # type: ignore
            data = message.choices[0].get("text", "")  # type: ignore
            yield data  # type: ignore


class AdaEmbeddings(OpenAIResource):
    """OpenAI Embeddings API.
    ---
    Ada is a transformer-based language model that can be used to generate embeddings for text. It is trained on a variety of text sources, including books, scientific papers, and reddit comments. You can read more about it in the [OpenAI blog post](https://openai.com/blog/new-and-improved-embedding-model).
    """

    model: EmbeddingModel = Field(default="text-embedding-ada-002")

    @handle
    async def run(self, texts: List[str]) -> List[Vector]:  # type: ignore
        """

        Given a list of texts, decorated with the `handle` utility, this method safely makes calls to the OpenAI Embeddings API with the appropiate format for a list of texts.
        It's purpose is to convert a list of texts into a list of vectors that can be consumed by the vector database for further implementation of the `query` and `upsert` methods of the `VectorClient` class regarding the `retrieval augmented generation` use case.

        Args:
            texts (List[str]): The list of texts.

        Returns:

                List[Vector]: The list of vectors.

        """

        response = await openai.Embedding.acreate(input=texts, model=self.model)  # type: ignore
        return [r.embedding for r in response.data]  # type: ignore


class DallE(OpenAIResource):
    """OpenAI Image API.
    ---
    Dall-E is a neural network that generates images from text descriptions, for example: "an illustration of a baby daikon radish in a tutu walking a dog". While it can be used for image generation in general, it was trained on text paired with images, so it works best when the text describes something in the image. You can read more about it in the [OpenAI blog post](https://openai.com/blog/dall-e/).
    """

    model: ImageModel = Field(default="dall-e")
    size: Size = Field(default="1024x1024")
    format: ImageFormat = Field(default="url")

    @handle
    async def run(self, text: str, n: int = 1) -> List[str]:  # type: ignore
        """

        Given an input text, decorated with the `handle` utility, this method safely makes calls to the OpenAI Image API with the appropiate format to generate a set of images from the input text.

        Args:

            text (str): The input text.

            n (int, optional): The number of images to generate. Defaults to 1.

        Returns:

            List[str]: The list of images.

        """

        response = await openai.Image.acreate(prompt=text, n=n, size=self.size, response_format=self.format)  # type: ignore
        return [r[self.format] for r in response.data]  # type: ignore


class Whisper(OpenAIResource):
    """OpenAI Audio API.
    ---
    Whisper is a neural network that converts text to speech. It is trained on a variety of English language audio sources, including audiobooks, podcasts, and YouTube videos. You can read more about it in the [OpenAI blog post](https://openai.com/research/whisper).
    """

    model: AudioModel = Field(default="whisper-1")

    @handle
    async def run(self, content: bytes, audioformat: AudioFormat = "wav") -> str:  # type: ignore
        """

        Given an input text, decorated with the `handle` utility, this method safely makes calls to the OpenAI Audio API with the appropiate format to infer the text from the input audio.

        Args:

            content (bytes): The input audio.

            audioformat (AudioFormat, optional): The audio format. Defaults to "wav".

        Returns:

            str: The inferred text.

        """

        response = await openai.Audio.acreate(self.model, AudioRequest(file=content, format=audioformat)())  # type: ignore
        return response.get("text", "")  # type: ignore


class VectorClient(APIClient):
    """

    A client for the vector database.

    Args:

        base_url (str, optional): The base url of the vector database. Defaults to "https://api.pinecone.io".

        headers (Dict[str, str], optional): The headers to use for the requests. Defaults to {"api-key": os.getenv("PINECONE_API_KEY"), "Content-Type": "application/json", "Accept": "application/json"}.

    """

    base_url: str = Field(
        default_factory=lambda: os.getenv("PINECONE_URL", "https://api.pinecone.io")
    )
    headers: Dict[str, str] = Field(
        default_factory=lambda: {
            "api-key": os.getenv("PINECONE_API_KEY"),
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
    )

    @property
    def embeddings(self) -> AdaEmbeddings:
        """

        Returns an instance of the `AdaEmbeddings` class to consume its methods.

        Returns:

            AdaEmbeddings: An instance of the `AdaEmbeddings` class.

        """
        return AdaEmbeddings()

    @handle
    async def upsert(self, embeddings: List[Embedding]) -> UpsertResponse:
        """
        Upserts embeddings into the vector database.

        Args:

                embeddings (List[Embedding]): The list of embeddings to upsert.

        Returns:

            UpsertResponse: The response object from the vector database.

        """
        response = await self.post(
            "/vectors/upsert",
            data={
                "vectors": [
                    UpsertRequest(
                        values=embedding.values, metadata=embedding.metadata
                    ).dict()
                    for embedding in embeddings
                ]
            },
        )
        assert response is not None, "Response is None"
        return UpsertResponse(**response)

    @handle
    async def query(
        self, expr: Query, vector: Vector, topK: int, includeMetadata: bool = True
    ) -> QueryResponse:
        """

        Queries the vector database.

        Args:

            expr (Query): The query expression.

            vector (Vector): The vector to query.

            topK (int): The number of results to return.

            includeMetadata (bool, optional): Whether to include metadata in the response. Defaults to True.

        Returns:

            QueryResponse: The response object from the vector database.

        """

        payload = QueryRequest(
            topK=topK,
            filter=expr,  # type: ignore
            vector=vector,
            includeMetadata=includeMetadata,
        ).dict()
        response = await self.post(
            "/query",
            data=payload,
        )
        assert response is not None, "Response is None"
        return QueryResponse(**response)

    @handle
    async def search(self, text: str, namespace: str, top_k: int = 5) -> List[str]:
        """Search for a text into a namespace

        Args:
            text (str): The text to search
            namespace (str): The namespace to search into
            top_k (int, optional): The number of results to return. Defaults to 5.

        Returns:
            List[str]: The list of results
        """
        response = await self.query(
            expr=(QueryBuilder("namespace") == namespace).query,
            vector=(await self.embeddings.run([text]))[0],
            topK=top_k,
            includeMetadata=True,
        )

        return [match.metadata["text"] for match in response.matches]  # type: ignore

    @handle
    async def run(
        self,
        text: str,
        namespace: str,
        vector: Vector,
        action: Literal["query", "upsert"],
    ):
        """Run a query or upsert a vector

        Args:
            text (str): The text to search
            namespace (str): The namespace to search into
            vector (Vector): The vector to upsert
            action (Literal["query", "upsert"]): The action to perform

        Raises:
            NotImplementedError: When an action that is not implemented is requested

        Returns:

            Union[List[str], int]: The list of results or the number of upserted vectors
        """

        if action == "query":
            response = await self.query(
                expr=(QueryBuilder("namespace") == namespace).query,
                vector=vector,
                topK=5,
                includeMetadata=True,
            )
            return [match.metadata["text"] for match in response.matches]  # type: ignore
        if action == "upsert":
            response = await self.upsert(
                [
                    Embedding(
                        values=vector, metadata={"text": text, "namespace": namespace}
                    )
                ]
            )
            return response.upsertedCount
        else:
            raise NotImplementedError("Action not implemented")



"""Automatically generate OpenAPI documentation for your Cloudantic API."""
from inspect import Parameter
from json import JSONDecoder, loads
from typing import Any, Dict, List

from aiohttp.web import Request
from aiohttp.web_request import FileField
from pydantic import BaseModel  # pylint: disable=no-name-in-module

from .json import parse_json_hook as parse_json


def extract(params: Dict[str, Parameter], path: str):
    """
    Extract OpenAPI parameters from the function parameters.

    Args:
        params (dict): The parameters of the function.
        path (str): The URL path of the endpoint.

    Returns:
        Dict[str, Any]: The extracted OpenAPI parameters.
    """
    open_api_params: Dict[str, Any] = {}

    for name, param in params.items():
        type_ = param.annotation

        if type_ in (str, int, float, bool) and name:
            param_location = "path" if f"{{{name}}}" in path else "query"

            open_api_params[name] = {
                "in": param_location,
                "name": name,
                "required": True,
                "schema": {"type": type_},
            }

            if param.default is not param.empty:
                open_api_params[name]["schema"]["default"] = param.default

        elif type_ in [FileField]:
            open_api_params[name] = {
                "in": "formData",
                "name": name,
                "required": True,
                "schema": {"type": "file"},
                "consumes": "multipart/form-data",
            }

        elif issubclass(type_, (BaseModel)):
            open_api_params[name] = {
                "in": "body",
                "name": name,
                "required": True,
                "schema": type_.schema(),
            }
        else:
            continue

    return open_api_params


def transform(
    open_api: Dict[str, Any],
    path: str,
    method: str,
    func: Any,
    open_api_params: Dict[str, Any],
    tags: List[str] = [],
):
    """
    Update the OpenAPI documentation with the endpoint information.

    Args:
        open_api (Dict[str, Any]): The OpenAPI documentation.
        path (str): The URL path of the endpoint.
        method (str): The HTTP method of the endpoint.
        func (Any): The function being documented.
        open_api_params (Dict[str, Any]): The OpenAPI parameters of the function.
    """
    if path in ["/openapi.json", "/docs"]:
        return

    _scalars: List[Dict[str, Any]] = []
    _body = None
    _is_file_upload = False

    for param in open_api_params.values():
        if isinstance(param["schema"], dict):
            if "type" in param["schema"] and param["schema"]["type"] != "object":
                _scalars.append(param)
            else:
                _body = {"content": {"application/json": {"schema": param["schema"]}}}
        elif param["in"] == "formData" and param["schema"]["type"] == "file":
            _is_file_upload = True
            _scalars.append(param)
        else:
            continue

    route_info = {
        "summary": func.__name__,
        "description": func.__doc__,
        "parameters": _scalars,
        "responses": {"200": {"description": "OK"}},
    }

    if tags:
        route_info["tags"] = tags

    if _body:
        open_api["paths"].setdefault(path, {})[method.lower()] = {
            "summary": func.__name__,
            "description": func.__doc__,
            "parameters": _scalars,
            "requestBody": _body
            if not _is_file_upload
            else {
                "content": {
                    "multipart/form-data": {
                        "schema": {
                            "type": "object",
                            "properties": {
                                "file": {"type": "string", "format": "binary"}
                            },
                        }
                    }
                }
            },
            "responses": {"200": {"description": "OK"}},
        }

    else:
        open_api["paths"].setdefault(path, {})[method.lower()] = {
            "summary": func.__name__,
            "description": func.__doc__,
            "parameters": _scalars,
            "responses": {"200": {"description": "OK"}},
        }


async def load(request: Request, params: Dict[str, Parameter]):
    """
    Shape the endpoint function parameters to match the request.

    Args:
        request (Request): The HTTP request.
        params (dict): The parameters of the function.

    Returns:
        Dict[str, Any]: The updated parameters to apply to the function.
    """
    args_to_apply: Dict[str, Any] = {}

    for name, param in params.items():
        annotation = param.annotation
        if annotation in (str, int, float, bool) and name in request.match_info:
            args_to_apply[name] = request.match_info[name]
        elif annotation in (str, int, float, bool) and name in request.query:
            args_to_apply[name] = annotation(request.query[name])
        elif annotation == FileField:
            file = (await request.post()).get("file")
            assert isinstance(file, FileField), "File not found"
            args_to_apply[name] = file
        elif issubclass(annotation, BaseModel):
            data = await request.json(loads=JSONDecoder().decode)
            if isinstance(data, (str, bytes)):
                data = loads(data, object_hook=parse_json)  # type: ignore
            args_to_apply[name] = annotation(**data)
        else:
            continue
    return args_to_apply


def html_string(base_url: str):
    """
    Returns a string with the html code to render the swagger ui
    """
    return f"""<!DOCTYPE html>
            <html lang="en">
            <head>
                <meta charset="UTF-8">
                <title>Cloudantic</title>
                <link rel="stylesheet" type="text/css" href="https://cdn.jsdelivr.net/npm/swagger-ui-dist@5/swagger-ui.css" >
                <link rel="icon" type="image/png" href="https://obahamonde.netlify.app/logo.png" sizes="32x32" />
                <style>
                .swagger-ui .topbar {{
                    display: none;
                }}
                </style>
            </head>
            <body>
                <div id="swagger-ui"></div>
                <script src="https://cdn.jsdelivr.net/npm/swagger-ui-dist@5/swagger-ui-bundle.js"> </script>
                <script src="https://cdn.jsdelivr.net/npm/swagger-ui-dist@5/swagger-ui-standalone-preset.js"> </script>
                <script>
                window.onload = function() {{
                const ui = SwaggerUIBundle({{
                    url: "{base_url}",
                    dom_id: '#swagger-ui',
                    deepLinking: true,
                    presets: [
                    SwaggerUIBundle.presets.apis,
                    SwaggerUIStandalonePreset
                    ],
                    plugins: [
                    SwaggerUIBundle.plugins.DownloadUrl
                    ],
                    layout: "StandaloneLayout"
                }})
                window.ui = ui
                }}
            </script>
            </body>
            </html>
            """



from typing import Any, Optional

import aiohttp
from pydantic import BaseModel, Field  # pylint: disable=no-name-in-module

from .utils import retry  # pylint: disable=relative-beyond-top-level
from .utils import setup_logging  # pylint: disable=relative-beyond-top-level

logger = setup_logging(__name__)


class APIClient(BaseModel):
    """
    A class representing an API client.

    Attributes:
    -----------
    headers : dict
        The headers to be used for the API requests.
    base_url : strÍ
        The base URL for the API requests.
    """

    base_url: str = Field(..., description="The base URL for the API requests")
    headers: dict[str, str] = Field(
        ..., description="The headers to be used for the API requests"
    )

    @retry()
    async def get(
        self,
        endpoint: str,
        params: Optional[dict[str, Any]] = None,
        session: Optional[aiohttp.ClientSession] = None,
    ):
        """
        Sends a GET request to the API.

        Parameters:
        -----------
        endpoint : str
            The endpoint to send the request to.
        params : Optional[dict[str, Any]]
            The query parameters to be used for the request.

        Returns:
        --------
        dict
            The response from the API.
        """
        if session is not None:
            try:
                response = await session.get(endpoint, params=params)
                data = await response.json()
                logger.info("GET %s %s %s", endpoint, response.status, data)
                return data
            except Exception as e:
                logger.error("GET %s %s", endpoint, e)
                raise e
        else:
            try:
                async with aiohttp.ClientSession(
                    base_url=self.base_url, headers=self.headers
                ) as session:
                    response = await session.get(endpoint, params=params)
                    data = await response.json()
                    logger.info("GET %s %s %s", endpoint, response.status, data)
                    return data
            except Exception as e:
                logger.error("GET %s %s", endpoint, e)
                raise e

    @retry()
    async def post(
        self,
        endpoint: str,
        data: Optional[dict[str, Any]] = None,
        session: Optional[aiohttp.ClientSession] = None,
    ):
        """
        Sends a POST request to the API.

        Parameters:
        -----------
        endpoint : str
            The endpoint to send the request to.
        data : Optional[dict[str, Any]]
            The data to be sent with the request.

        Returns:
        --------
        dict
            The response from the API.
        """
        if session is not None:
            try:
                response = await session.post(endpoint, json=data)
                data = await response.json()
                logger.info("POST %s %s %s", endpoint, response.status, data)
                return data
            except Exception as e:
                logger.error("POST %s %s", endpoint, e)
                raise e
        else:
            try:
                async with aiohttp.ClientSession(
                    base_url=self.base_url, headers=self.headers
                ) as session:
                    response = await session.post(endpoint, json=data)
                    data = await response.json()
                    logger.info("POST %s %s %s", endpoint, response.status, data)
                    return data
            except Exception as e:
                logger.error("POST %s %s", endpoint, e)
                raise e

    @retry()
    async def put(
        self,
        endpoint: str,
        data: Optional[dict[str, Any]] = None,
        session: Optional[aiohttp.ClientSession] = None,
    ):
        """
        Sends a PUT request to the API.

        Parameters:
        -----------
        endpoint : str
            The endpoint to send the request to.
        data : Optional[dict[str, Any]]
            The data to be sent with the request.

        Returns:
        --------
        dict
            The response from the API.
        """
        if session is not None:
            try:
                response = await session.put(endpoint, json=data)
                data = await response.json()
                logger.info("PUT %s %s %s", endpoint, response.status, data)
                return data
            except Exception as e:
                logger.error("PUT %s %s", endpoint, e)
                raise e
        else:
            try:
                async with aiohttp.ClientSession(
                    base_url=self.base_url, headers=self.headers
                ) as session:
                    response = await session.put(endpoint, json=data)
                    data = await response.json()
                    logger.info("PUT %s %s %s", endpoint, response.status, data)
                    return data
            except Exception as e:
                logger.error("PUT %s %s", endpoint, e)
                raise e

    @retry()
    async def delete(
        self,
        endpoint: str,
        params: Optional[dict[str, Any]] = None,
        session: Optional[aiohttp.ClientSession] = None,
    ):
        """
        Sends a DELETE request to the API.

        Parameters:
        -----------
        endpoint : str
            The endpoint to send the request to.

        Returns:
        --------
        dict
            The response from the API.
        """
        if session is not None:
            try:
                response = await session.delete(endpoint, params=params)
                data = await response.json()
                logger.info("DELETE %s %s %s", endpoint, response.status, data)
                return data
            except Exception as e:
                logger.error("DELETE %s %s", endpoint, e)
                raise e
        else:
            async with aiohttp.ClientSession(
                base_url=self.base_url, headers=self.headers
            ) as session:
                try:
                    response = await session.delete(endpoint, params=params)
                    data = await response.json()
                    logger.info("DELETE %s %s %s", endpoint, response.status, data)
                    return data
                except Exception as e:
                    logger.error("DELETE %s %s", endpoint, e)
                    raise e

    @retry()
    async def patch(
        self,
        endpoint: str,
        data: Optional[dict[str, Any]] = None,
        session: Optional[aiohttp.ClientSession] = None,
    ):
        """
        Sends a PATCH request to the API.
        """
        if session is not None:
            try:
                response = await session.patch(endpoint, json=data)
                data = await response.json()
                logger.info("PATCH %s %s %s", endpoint, response.status, data)
                return data
            except Exception as e:
                logger.error("PATCH %s %s", endpoint, e)
                raise e
        else:
            async with aiohttp.ClientSession(
                base_url=self.base_url, headers=self.headers
            ) as session:
                try:
                    response = await session.patch(endpoint, json=data)
                    data = await response.json()
                    logger.info("PATCH %s %s %s", endpoint, response.status, data)
                    return data
                except Exception as e:
                    logger.error("PATCH %s %s", endpoint, e)
                    raise e

    @retry()
    async def head(
        self, endpoint: str, session: Optional[aiohttp.ClientSession] = None
    ):
        """
        Sends a HEAD request to the API.

        Parameters:
        -----------
        endpoint : str
            The endpoint to send the request to.

        Returns:
        --------
        dict
            The response from the API.
        """
        if session is not None:
            try:
                response = await session.head(endpoint)
                data = await response.json()
                logger.info("HEAD %s %s %s", endpoint, response.status, data)
                return data
            except Exception as e:
                logger.error("HEAD %s %s", endpoint, e)
                raise e
        else:
            async with aiohttp.ClientSession(
                base_url=self.base_url, headers=self.headers
            ) as session:
                try:
                    response = await session.head(endpoint)
                    data = await response.json()
                    logger.info("HEAD %s %s %s", endpoint, response.status, data)
                    return data
                except Exception as e:
                    logger.error("HEAD %s %s", endpoint, e)
                    raise e

    @retry()
    async def options(
        self, endpoint: str, session: Optional[aiohttp.ClientSession] = None
    ):
        """
        Sends an OPTIONS request to the API.

        Parameters:
        -----------
        endpoint : str
            The endpoint to send the request to.

        Returns:
        --------
        dict
            The response from the API.
        """
        if session is not None:
            try:
                response = await session.options(endpoint)
                data = await response.json()
                logger.info(f"OPTIONS {endpoint} {response.status} {data}")
                return data
            except Exception as e:
                logger.error("OPTIONS %s %s", endpoint, e)
                raise e
        else:
            async with aiohttp.ClientSession(
                base_url=self.base_url, headers=self.headers
            ) as session:
                try:
                    response = await session.options(endpoint)
                    data = await response.json()
                    logger.info(f"OPTIONS {endpoint} {response.status} {data}")
                    return data
                except Exception as e:
                    logger.error("OPTIONS %s %s", endpoint, e)
                    raise e

    @retry()
    async def text(
        self,
        endpoint: str,
        params: Optional[dict[str, Any]] = None,
        session: Optional[aiohttp.ClientSession] = None,
    ):
        """
        Sends a GET request to the API and returns the response as text.

        Parameters:
        -----------
        endpoint : str
            The endpoint to send the request to.

        Returns:
        --------
        str
            The response from the API.
        """
        if session is not None:
            try:
                response = await session.get(endpoint, params=params)
                data = response.text
                logger.info(f"GET {endpoint} {response.status} {data}")
                return data
            except Exception as e:
                logger.error(f"GET {endpoint} {e}")
                raise e
        else:
            async with aiohttp.ClientSession(
                base_url=self.base_url, headers=self.headers
            ) as session:
                try:
                    response = await session.get(endpoint, params=params)
                    data = response.text
                    logger.info(f"GET {endpoint} {response.status} {data}")
                    return data
                except Exception as e:
                    logger.error(f"GET {endpoint} {e}")
                    raise e

    @retry()
    async def blob(
        self,
        endpoint: str,
        params: Optional[dict[str, Any]] = None,
        session: Optional[aiohttp.ClientSession] = None,
    ):
        """
        Sends a GET request to the API and returns the response as bytes.

        Parameters:
        -----------
        endpoint : str
            The endpoint to send the request to.

        Returns:
        --------
        bytes
            The response from the API.
        """
        if session is not None:
            try:
                response = await session.get(endpoint, params=params)
                data = response.content
                logger.info(f"GET {endpoint} {response.status} {data}")
                return data
            except Exception as e:
                logger.error(f"GET {endpoint} {e}")
                raise e
        else:
            async with aiohttp.ClientSession(
                base_url=self.base_url, headers=self.headers
            ) as session:
                try:
                    response = await session.get(endpoint, params=params)
                    data = response.content
                    logger.info(f"GET {endpoint} {response.status} {data}")
                    return data
                except Exception as e:
                    logger.error(f"GET {endpoint} {e}")
                    raise e

    async def stream(
        self,
        endpoint: str,
        params: Optional[dict[str, Any]],
        session: Optional[aiohttp.ClientSession] = None,
    ):
        """
        Sends a GET request to the API and returns the response as a stream.

        Parameters:
        -----------
        endpoint : str
            The endpoint to send the request to.

        Returns:
        --------
        httpx.Response
            The response from the API.
        """
        if session is not None:
            try:
                async with session.request("GET", endpoint, params=params) as response:
                    async for chunk in response.content.iter_any():
                        yield chunk
            except Exception as e:
                logger.error(f"GET {endpoint} {e}")
                raise e
        else:
            async with aiohttp.ClientSession(
                base_url=self.base_url, headers=self.headers
            ) as session:
                try:
                    async with session.request(
                        "GET", endpoint, params=params
                    ) as response:
                        async for chunk in response.content.iter_any():
                            yield chunk
                except Exception as e:
                    logger.error(f"GET {endpoint} {e}")
                    raise e



"""JSON utilities for liferay_llm."""
import base64
from datetime import date, datetime, time
from decimal import Decimal
from enum import Enum
from json import JSONEncoder
from typing import Any, Dict
from uuid import UUID

from pydantic import BaseModel  # pylint: disable=no-name-in-module


class CloudEncoder(JSONEncoder):
    """
    Encoder for:
    -----------
    - datetime.datetime
    - uuid.UUID
    - float
    """

    def default(self, o: Any):
        if isinstance(o, datetime):
            return o.isoformat()
        if isinstance(o, UUID):
            return str(o)
        if isinstance(o, float):
            return Decimal(str(o))
        if isinstance(o, BaseModel):
            return o.dict()
        return super().default(o)


def parse_json_hook(dct: Dict[str, Any]):
    """
    Cast float to Decimal.
    """
    for k, v in dct.items():
        if isinstance(v, float):
            dct[k] = Decimal(str(v))

    return dct


def to_json(data: Any) -> str:
    """
    Convert data to JSON.
    """
    return CloudEncoder().encode(data)


def jsonable_encoder(
    obj: Any,
    *,
    include: list[str] = [],
    exclude: list[str] = [],
    by_alias: bool = False,
    skip_defaults: bool = False,
    custom_encoder: Any = None,
) -> Any:
    """
    Convert any object to a JSON-serializable object.

    This function is used by Cloudantic to convert objects to JSON-serializable objects.

    It supports all the types supported by the standard json library, plus:

    * datetime.datetime
    * datetime.date
    * datetime.time
    * uuid.UUID
    * enum.Enum
    * pydantic.BaseModel
    """

    if custom_encoder is None:
        custom_encoder = CloudEncoder

    if obj is str:
        return "string"
    if obj is int or obj is float:
        return "integer"
    if obj is bool:
        return "boolean"
    if obj is None:
        return "null"
    if obj is list:
        return "array"
    if obj is dict:
        return "object"
    if obj is bytes:
        return "binary"
    if obj is datetime:
        return "date-time"
    if obj is date:
        return "date"
    if obj is time:
        return "time"
    if obj is UUID:
        return "uuid"
    if obj is Enum:
        return "enum"
    if isinstance(obj, (str, int, float, bool, type(None))):
        return obj
    if isinstance(obj, (list, tuple, set, frozenset)):
        return [
            jsonable_encoder(
                v,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                skip_defaults=skip_defaults,
                custom_encoder=custom_encoder,
            )
            for v in obj  # type: ignore
        ]
    if isinstance(obj, dict):
        return {
            jsonable_encoder(
                k,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                skip_defaults=skip_defaults,
                custom_encoder=custom_encoder,
            ): jsonable_encoder(
                v,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                skip_defaults=skip_defaults,
                custom_encoder=custom_encoder,
            )
            for k, v in obj.items()  # type: ignore
        }
    if isinstance(obj, bytes):
        return base64.b64encode(obj).decode()
    if isinstance(obj, (set, frozenset)):
        return [
            jsonable_encoder(
                v,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                skip_defaults=skip_defaults,
                custom_encoder=custom_encoder,
            )
            for v in obj  # type: ignore
        ]
    if isinstance(obj, datetime):
        return obj.isoformat()
    if isinstance(obj, Enum):
        return obj.value
    if isinstance(obj, UUID):
        return str(obj)
    if isinstance(obj, type):
        return jsonable_encoder(
            obj.__name__,
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            custom_encoder=custom_encoder,
        )

    return custom_encoder().default(obj)



"""
Helper functions for Cloudantic.
"""
from functools import singledispatch
from typing import Any

from aiohttp.web import Response, json_response
from pydantic import BaseModel  # pylint: disable=no-name-in-module

from .json import to_json  # pylint: disable=no-name-in-module


@singledispatch
def do_response(response: Any) -> Response:
    """
    Process the response from a view function and return an aiohttp.web.Response object.
    """

    return response


@do_response.register(BaseModel)
def _(response: BaseModel) -> Response:
    return json_response(response.dict(exclude_none=True), dumps=to_json)


@do_response.register(dict)
def _(response: dict[str, Any]) -> Response:
    return json_response(response, dumps=to_json)


@do_response.register(str)
def _(response: str) -> Response:
    return Response(status=200, text=response, content_type="text/html")


@do_response.register(bytes)
def _(response: bytes) -> Response:
    return Response(
        status=200, body=response, content_type="application/octet-sse_stream"
    )


@do_response.register(int)
def _(response: int) -> Response:
    return Response(status=200, text=str(response), content_type="text/plain")


@do_response.register(float)
def _(response: float) -> Response:
    return Response(status=200, text=str(response), content_type="text/plain")


@do_response.register(bool)
def _(response: bool) -> Response:
    return Response(status=200, text=str(response), content_type="text/plain")


@do_response.register(tuple)
def _(response: tuple[Any]) -> Response:
    return do_response(list(response))


@do_response.register(set)
def _(response: set[Any]) -> Response:
    return do_response(list(response))


@do_response.register(frozenset)
def _(response: frozenset[Any]) -> Response:
    return do_response(list(response))


@do_response.register(type(None))
def _(response: None) -> Response:
    return Response(status=200, text="", content_type="text/plain")


@do_response.register(Response)
def _(response: Response) -> Response:
    return response



from __future__ import annotations

from tempfile import NamedTemporaryFile
from typing import Any, Dict, List, Literal, Set, Type, TypeAlias, TypeVar, Union
from uuid import uuid4

from pydantic import Field  # pylint: disable=no-name-in-module
from pydantic import BaseModel, root_validator

from .typedef import (
    AudioFormat,
    ChatModel,
    CompletionModel,
    ImageFormat,
    MetaData,
    Role,
    Size,
    Vector,
)

# pylint: disable=no-name-in-module


Value: TypeAlias = Union[int, float, str, bool, List[str]]
QueryJoin: TypeAlias = Literal["$and", "$or"]
QueryWhere: TypeAlias = Literal[
    "$eq", "$ne", "$gt", "$gte", "$lt", "$lte", "$in", "$nin"
]
QueryKey: TypeAlias = Union[str, QueryWhere, QueryJoin]
QueryValue: TypeAlias = Union[Value, List[Value], "Query", List["Query"]]
Query: TypeAlias = Dict[QueryKey, QueryValue]
# Please write Sphinx documentation for this m

T = TypeVar("T")


class QueryBuilder(object):
    def __init__(self, field: str = None, query: Query = None):  # type: ignore
        """
        Initializes a new QueryBuilder instance.

        Args:
            field (str, optional): The field to query. Defaults to None.
            query (Query, optional): The query dictionary. Defaults to an empty dictionary.
        """
        self.field = field
        self.query = query if query else {}

    def __repr__(self) -> str:
        """Returns the string representation of the query."""
        return f"{self.query}"

    def __call__(self, field_name: str) -> QueryBuilder:
        """
        Creates a new QueryBuilder instance with a given field name.

        Args:
            field_name (str): The field name to query.

        Returns:
            QueryBuilder: A new QueryBuilder instance.
        """
        return QueryBuilder(field_name)

    def __and__(self, other: QueryBuilder) -> QueryBuilder:
        """
        Combines two queries using the $and operator.

        Args:
            other (QueryBuilder): Another QueryBuilder instance.

        Returns:
            QueryBuilder: A new QueryBuilder instance with the combined query.
        """
        return QueryBuilder(query={"$and": [self.query, other.query]})

    def __or__(self, other: QueryBuilder) -> QueryBuilder:
        """
        Combines two queries using the $or operator.

        Args:
            other (QueryBuilder): Another QueryBuilder instance.

        Returns:
            QueryBuilder: A new QueryBuilder instance with the combined query.
        """
        return QueryBuilder(query={"$or": [self.query, other.query]})

    def __eq__(self, value: Value) -> QueryBuilder:  # type: ignore
        """
        Creates a new QueryBuilder instance with the $eq operator.

        Args:
            value (Value): The value to query.

        Returns:
            QueryBuilder: A new QueryBuilder instance with the combined query.

        """
        return QueryBuilder(query={self.field: {"$eq": value}})

    def __ne__(self, value: Value) -> QueryBuilder:  # type: ignore
        """
        Creates a new QueryBuilder instance with the $ne operator.

        Args:
            value (Value): The value to query.

        Returns:
            QueryBuilder: A new QueryBuilder instance with the combined query.

        """
        return QueryBuilder(query={self.field: {"$ne": value}})

    def __lt__(self, value: Value) -> QueryBuilder:
        """
        Creates a new QueryBuilder instance with the $lt operator.

        Args:

            value (Value): The value to query.

        Returns:

            QueryBuilder: A new QueryBuilder instance with the combined query.


        """

        return QueryBuilder(query={self.field: {"$lt": value}})

    def __le__(self, value: Value) -> QueryBuilder:
        """
        Creates a new QueryBuilder instance with the $lte operator.

        Args:

            value (Value): The value to query.

        Returns:

            QueryBuilder: A new QueryBuilder instance with the combined query.

        """

        return QueryBuilder(query={self.field: {"$lte": value}})

    def __gt__(self, value: Value) -> QueryBuilder:
        """

        Creates a new QueryBuilder instance with the $gt operator.

        Args:

            value (Value): The value to query.

        Returns:

            QueryBuilder: A new QueryBuilder instance with the combined query.

        """

        return QueryBuilder(query={self.field: {"$gt": value}})

    def __ge__(self, value: Value) -> QueryBuilder:
        """

        Creates a new QueryBuilder instance with the $gte operator.

        Args:

            value (Value): The value to query.

        Returns:

            QueryBuilder: A new QueryBuilder instance with the combined query.

        """

        return QueryBuilder(query={self.field: {"$gte": value}})

    def in_(self, values: List[Value]) -> QueryBuilder:
        """

        Creates a new QueryBuilder instance with the $in operator.

        Args:

            values (List[Value]): The values to query.

        Returns:

            QueryBuilder: A new QueryBuilder instance with the combined query.

        """

        return QueryBuilder(query={self.field: {"$in": values}})

    def nin_(self, values: List[Value]) -> QueryBuilder:
        """

        Creates a new QueryBuilder instance with the $nin operator.

        Args:

            values (List[Value]): The values to query.

        Returns:

            QueryBuilder: A new QueryBuilder instance with the combined query.

        """

        return QueryBuilder(query={self.field: {"$nin": values}})


class UpsertRequest(BaseModel):
    """
    A request to upsert a vector.

    Args:
        id (str, optional): The ID of the vector. Defaults to a random UUID.
        values (Vector): The vector values.
        metadata (MetaData): The vector metadata.
    """

    id: str = Field(default_factory=lambda: str(uuid4()))
    values: Vector = Field(...)
    metadata: MetaData = Field(...)


class Embedding(BaseModel):
    """

    A vector embedding.

    Args:

        values (Vector): The vector values.

        metadata (MetaData): The vector metadata.

    """

    values: Vector = Field(...)
    metadata: MetaData = Field(...)


class QueryRequest(BaseModel):
    """
    A request to query vectors.

    Args:

        topK (int, optional): The number of results to return. Defaults to 10.

        filter (Query): The query filter.

        includeMetadata (bool, optional): Whether to include metadata in the response. Defaults to True.

        vector (Vector): The vector to query.

    """

    topK: int = Field(default=10)
    filter: Dict[str, Any] = Field(...)
    includeMetadata: bool = Field(default=True)
    vector: Vector = Field(...)


class QueryMatch(BaseModel):
    """

    A query match.

    Args:

        id (str): The ID of the vector.

        score (float): The match score.

        metadata (MetaData): The vector metadata.

    """

    id: str = Field(...)
    score: float = Field(...)
    metadata: MetaData = Field(...)


class QueryResponse(BaseModel):
    """

    A response to a query request.

    Args:

        matches (List[QueryMatch]): The query matches.

    """

    matches: List[QueryMatch] = Field(...)


class UpsertResponse(BaseModel):
    """

    A response to an upsert request.

    Args:

        upsertedCount (int): The number of vectors upserted.

    """

    upsertedCount: int = Field(...)


class CompletionRequest(BaseModel):
    model: CompletionModel = Field(default="gpt-3.5-turbo-instruct")
    prompt: str = Field(...)
    temperature: float = Field(default=0.2)
    max_tokens: int = Field(default=1024)
    stream: bool = Field(default=False)


class CompletionChoice(BaseModel):
    index: int = Field(...)
    finish_reason: str = Field(...)
    text: str = Field(...)


class CompletionUsage(BaseModel):
    prompt_tokens: int = Field(...)
    completion_tokens: int = Field(...)
    total_tokens: int = Field(...)


class CompletionResponse(BaseModel):
    id: str = Field(...)
    object: str = Field(...)
    created: int = Field(...)
    model: CompletionModel = Field(...)
    choices: List[CompletionChoice] = Field(...)
    usage: CompletionUsage = Field(...)


class Message(BaseModel):
    """
    Represents a message within the chatcompletion API pipeline.
    """

    role: Role = Field(default="user")
    content: str = Field(...)


class ChatCompletionRequest(BaseModel):
    model: ChatModel = Field(default="gpt-3.5-turbo-16k")
    messages: List[Message] = Field(...)
    temperature: float = Field(default=0.5)
    max_tokens: int = Field(default=512)
    stream: bool = Field(default=False)


class ChatCompletionUsage(BaseModel):
    """Token usage statistics for a chat completion API call."""

    prompt_tokens: int = Field(...)
    completion_tokens: int = Field(...)
    total_tokens: int = Field(...)


class ChatCompletionChoice(BaseModel):
    index: int = Field(...)
    message: Message = Field(...)
    finish_reason: str = Field(...)


class ChatCompletionResponse(BaseModel):
    id: str = Field(...)
    object: str = Field(...)
    created: int = Field(...)
    model: str = Field(...)
    choices: List[ChatCompletionChoice] = Field(...)
    usage: ChatCompletionUsage = Field(...)


class EmbeddingUssage(BaseModel):
    """Token usage statistics for an embedding API call."""

    prompt_tokens: int = Field(...)
    total_tokens: int = Field(...)


class CreateImageResponse(BaseModel):
    created: float = Field(...)
    data: List[Dict[ImageFormat, str]] = Field(...)


class CreateImageRequest(BaseModel):
    """Request to create an image from a prompt. Use default values for configuration unless specified."""

    prompt: str = Field(...)
    n: int = Field(default=1)
    size: Size = Field(default="1024x1024")
    response_format: ImageFormat = Field(default="url")


class FineTuneSample(BaseModel):
    messages: List[Message] = Field(..., max_items=3, min_items=2)

    @root_validator
    @classmethod
    def check_messages(cls: Type[BaseModel], values: Dict[str, Any]):
        roles: Set[Role] = set()
        for message in values["messages"]:
            roles.add(message.role)
        assert len(roles) == len(
            values["messages"]
        ), "All messages must be from different roles."
        return values


class FineTuneRequest(BaseModel):
    __root__: List[FineTuneSample] = Field(..., min_items=10, max_items=100000)

    def __call__(self):
        with NamedTemporaryFile("w", suffix=".json") as f:
            data = self.json()
            f.write(data)
            f.flush()
            return f


class AudioRequest(BaseModel):
    file: bytes = Field(...)
    format: AudioFormat = Field(default="mp3")

    def __call__(self):
        with NamedTemporaryFile("wb", suffix=f".{self.format}") as f:
            f.write(self.file)
            f.flush()
            assert len(f.read()) < 25 * 1024 * 1024, "File too large."
            return f


class AudioRequest(BaseModel):
    file: bytes = Field(...)
    format: AudioFormat = Field(default="mp3")

    def __call__(self):
        with NamedTemporaryFile("wb", suffix=f".{self.format}") as f:
            f.write(self.file)
            f.flush()
            assert len(f.read()) < 25 * 1024 * 1024, "File too large."
            return f
